unit chain2;
interface
const DefGranularity=32;
type
     aCharVector=array[1..1] of char;
     aCharVector_=^aCharVector;
     Chain=object Body_: aCharVector_;
           Lo,Hi,Size,Granularity: word;
           procedure Init(st: string);
           procedure SetGranularity(gran: word);
           procedure Clear;
           function  Length: word;
           procedure Upcase;
           function  Substring(pos,lgth: word): string;
           function  Compare(var ch: chain): char;
	   procedure Delete(pos,lgth: word);
           procedure CopyString(source: string);
           procedure CopyChain(var source: Chain);
	   procedure ReInit(st: string);
           procedure AppendString(source: string);
           procedure AppendChain(var source: Chain);
           procedure PrependString(source: string);
           procedure PrependChain(var source: Chain);
           procedure InsertStringBefore(source: string; pos: word);
           procedure InsertChainBefore(var source: chain;  pos: word);
           procedure Collapse(ch: char);
           procedure CopySubchain(source: Chain; pos,lgth: word);
           function  PosChar(c: char): word;
           function  PosString(s: string):  word;
           function  StringFoundAt(s: string; pos: word):  boolean;
	   function  ChainFoundAt(s: chain; pos: word):  boolean;
           procedure FrontStrip(ch: char);
           procedure EndStrip(ch: char);
           procedure Read(terminator: char);
           procedure AppendInput(terminator: char);
           procedure Write;
           procedure Reverse;
           procedure Shrink;
           procedure ExpandTo(newLength: word);
     end;


function ChainResult: byte;
function ChainErrorMsg: string;

implementation

const Overflow=1; DiskSpace=2; Memory=3;
var ChainFail: byte;

function ChainResult: byte;
begin ChainResult:=ChainFail
end;


function ChainErrorMsg: string;
begin case ChainFail of
         0: ChainErrorMsg:='';
         1: ChainErrorMsg:='Maximum chain length reached';
         2: ChainErrorMsg:='Out of disk space';
         3: ChainErrorMsg:='Out of memory';
      end
end;

function  chain.Compare(var ch: chain): char;
var i,j: word; c: char;
begin
      i:=self.Lo-1; j:=ch.Lo-1; c:='=';
      while (i<self.Hi) and (j<ch.Hi) do
      begin inc(i); inc(j);
             if self.Body_^[i]<ch.Body_^[j] then c:='<' else
             if self.Body_^[i]>ch.Body_^[j] then c:='>'
      end;
      if c='=' then
      if self.Length<ch.Length then c:='<' else
      if self.Length>ch.Length then c:='>';
      Compare:=c
end;

procedure chain.Init(st: string);
begin Lo:=1; Hi:=system.length(st);
      Granularity:=DefGranularity;
      Size:=(Hi+Granularity-1) div Granularity*Granularity;
      if Size=0 then Body_:=nil
      else
      begin GetMem(Body_,Size); Move(st[1],Body_^,Hi);
      end;
end;

procedure chain.SetGranularity(gran: word);
var NewSize: word;
begin if (gran=Granularity) or (gran=0) then exit; (* nothing to do *);
      if Lo>1 then Shrink;
      NewSize:=(Hi+Gran-1) div Gran*Gran;
      Granularity:=Gran;
      if NewSize>Size then ExpandTo(NewSize);
end;


procedure chain.Clear;
begin Lo:=1; Hi:=0
end;

function chain.Length: word;
begin Length:=Hi+1-Lo
end;

procedure chain.Upcase;
var i: word;
begin for i:=Lo to Hi do Body_^[i]:=system.Upcase(Body_^[i]);
      (** using a pointer, C-style, does not improve performance **)
      (** the TP 5.5 compiler optimizes such loops **)
end;


function chain.Substring(pos,lgth: word): string;
var tmp: string;
begin
      if pos<=Hi then
      begin if pos+lgth-1>Hi then lgth:=Hi-pos+1;
            Move(Body_^[pos],tmp[1],lgth);
            tmp[0]:=char(lgth);
            Substring:=tmp;
      end
      else Substring:='';
end;

procedure chain.Delete(pos,lgth: word);
begin pos:=pos+Lo-1;
      if pos>Hi then exit;
      if pos+lgth-1>Hi then lgth:=Hi-pos+1;
      if pos=1 then inc(Lo,lgth) else
      begin dec(Hi,lgth);
         if pos+lgth-1<Hi then
         Move(Body_^[pos+lgth],Body_^[pos],Hi-lgth+1);
      end
end;

procedure chain.CopyString(source: string);
var lgth: word;
begin lgth:=system.length(source);
      if lgth<=Hi+1-Lo (* optimization of Length *)
      then
      begin Move(source[1],Body_^,lgth);
	    Lo:=1; Hi:=lgth;
      end
      else
      begin Lo:=1; Hi:=0;
	    AppendString(source);
      end
end;

procedure chain.CopyChain(var source: Chain);
var lgth: word;
begin with source do lgth:=Hi+1-Lo (* optimization of source.Length *) ;
      if lgth<=Hi+1-Lo (* optimization of Length *)
      then
      begin Move(Source.Body_^,Body_^,lgth); Lo:=1; Hi:=lgth;
      end
      else
      begin Lo:=1;
	    Hi:=0;
	    AppendChain(source);
      end
end;

procedure chain.ReInit(st: string);
begin FreeMem(Body_,Size); Init(st)
end;

procedure chain.AppendString(source: string);
var lgth: word;
begin lgth:=system.length(source);
      if Hi+lgth>Size then
      begin if Lo>1 then Shrink; ExpandTo(Size+lgth);
      end;
      Move(source[1],Body_^[Hi+1],lgth);
      inc(Hi,lgth)
end;

procedure chain.AppendChain(var source: Chain);
var lgth: word; i,J: word;
begin
      with source do lgth:=Hi+1-Lo (* optimization of source.Length *) ;
      if Hi+lgth>Size then
      begin if Lo>1 then Shrink; ExpandTo(Size+lgth);
      end;
      Move(source.Body_^[source.Lo],Body_^[Hi+1],lgth);
      inc(Hi,lgth);
end;

procedure chain.PrependString(source: string);
var lgth: word;
begin lgth:=system.length(source);
      if Hi+lgth-1>Size then
      begin if Lo>1 then Shrink; ExpandTo(Size+lgth);
      end;
      Move(Body_^[Lo],Body_^[Lo+lgth],Hi+1-Lo);
      Move(source[1],Body_^[Lo],lgth);
      inc(Hi,lgth)
end;

procedure chain.PrependChain(var source: Chain);
var lgth: word;
begin with source do lgth:=Hi+1-Lo (* optimization of lgth:=source.Length *) ;
      if Hi+lgth-1>Size then
      begin if Lo>1 then Shrink; ExpandTo(Size+lgth);
      end;
      Move(Body_^[Lo],Body_^[Lo+lgth],Hi+1-Lo (* optimization of Length *) );
      Move(source.Body_^[Lo],Body_^[Lo],lgth);
      inc(Hi,lgth)
end;

procedure chain.InsertStringBefore(source: string; pos: word);
var lgth: word;
begin if (pos>0) and (pos<=Hi) then
begin lgth:=system.length(source);
      if Hi+lgth-1>Size then
      begin if Lo>1 then Shrink; ExpandTo(Size+lgth);
      end;
      Move(Body_^[pos],Body_^[pos+lgth],Hi+1-Lo (* optimization of Length *) );
      Move(source[1],Body_^[pos],lgth);
      inc(Hi,lgth)
end;
end;

procedure chain.InsertChainBefore(var source: chain;  pos: word);
var lgth: word;
begin if (pos>0) and (pos<=Hi) then
begin with source do lgth:=Hi+1-Lo (* optimization of lgth:=source.Length *);
      if Hi+lgth-1>Size then
      begin if Lo>1 then Shrink; ExpandTo(Size+lgth);
      end;
      Move(Body_^[pos],Body_^[pos+lgth],
           Hi+1-Lo (* optimization of lgth:=source.Length *) );
      Move(source.Body_^[Lo],Body_^[pos],lgth);
      inc(Hi,lgth)
end;
end;

procedure chain.CopySubchain(source: Chain; pos,lgth: word);
begin Inc(pos,source.Lo-1);
      if pos>source.Hi then begin Lo:=1; Hi:=0 end
      else
      begin if pos+lgth>source.Hi then lgth:=source.Hi-pos;
      if lgth<=Hi+1-Lo (* optimization of Length *)
      then
      Move(Source.Body_^[pos],Body_^,lgth)
      else
      begin if Lo>1 then Shrink; ExpandTo(Size+lgth);
	    Move(source.Body_^[pos],Body_^[1],lgth);
      end;
      Lo:=1; Hi:=lgth
      end;
end;

function chain.PosChar(c: char): word;
var p: word;
begin p:=Lo; while (p<=Hi) and (Body_^[p]<>c) do inc(p);
      if Lo>Hi then PosChar:=0 else PosChar:=p
end;

function chain.PosString(s: string):  word;
var Start,Stop: word;  lgth: byte;
    ch_: ^char; ch: char;
    tmp: string;
    found:  boolean;
begin PosString:=0;
     if system.length(s)<=Hi+1-Lo (* optimization of Length *)
     then
begin lgth:=system.length(s); Stop:=Hi+1-lgth; tmp[0]:=char(lgth);
      Start:=Lo; ch_:=Addr(Body_^[Start]);
      Found:=false; ch:=s[1];
      repeat while (Start<Stop) and (ch_^<>ch)
             do begin inc(Start); inc(longint(ch_));
             end;
             Move(ch_^,tmp[1],lgth); found:=tmp=s;
	     if not found then
             begin inc(Start); inc(longint(ch_));
             end;
      until found or (Start>Stop);
      if found then PosString:=Start else PosString:=0;
end;
end;

function chain.StringFoundAt(s: string; pos: word):  boolean;
var lgth: byte;
    tmp: string;
    ch_,s_: ^char;
begin lgth:=system.length(s); StringFoundAt:=false;
      if  (pos+lgth>pos)   (* s<>'' and not beyond maximum chain length *)
      and (pos+lgth<=Hi+1) (* not beyond end of this chain *)
      then
      begin ch_:=Addr(Body_^[pos]); s_:=Addr(s[1]);
	    while (lgth>=1) and (ch_^=s_^) do
	    begin dec(lgth); inc(longint(ch_)); inc(longint(s_))
	    end;
	    StringFoundAt:=(lgth=0)
      end;
end;

function chain.ChainFoundAt(s: chain; pos: word):  boolean;
var lgth: word;
    tmp: string;
    ch_,s_: ^char;
begin with s do lgth:=Hi+1-Lo (* optimization of lgth:=s.Length *) ;
      ChainFoundAt:=false;
      if  (pos+lgth>pos)   (* s<>'' and not beyond maximum chain length *)
      and (pos+lgth<=Hi+1) (* not beyond end of this chain *)
      then
      begin ch_:=Addr(Body_^[pos]); s_:=Addr(s.Body_^[Lo]);
	    while (lgth>=1) and (ch_^=s_^) do
	    begin dec(lgth); inc(longint(ch_)); inc(longint(s_))
	    end;
	    ChainFoundAt:=(lgth=0)
      end;
end;

procedure chain.FrontStrip(ch: char);
begin while(Lo<=Hi) and (Body_^[Lo]=ch) do inc(Lo);
end;

procedure chain.EndStrip(ch: char);
begin while(Hi>=Lo) and (Body_^[Hi]=ch) do dec(Hi);
end;

procedure chain.Reverse;
var ch: char; lo_,hi_: ^char;
begin lo_:=Addr(Body_^[Lo]); hi_:=Addr(Body_^[Hi]);
      while longint(lo_)<longint(hi_) do
      begin ch:=lo_^; lo_^:=hi_^; hi_^:=ch;
            inc(longint(lo_)); dec(longint(hi_))
      end;
end;

procedure chain.Shrink;
begin Move(Body_^[Lo],Body_^[1],Hi+1-Lo (* optimization of Length *) );
      Dec(Hi,Lo-1); Lo:=1
end;

procedure chain.Collapse(ch: char);
(* not efficient, but probably a procedure seldom called *)
var chStart,p: word; c1,c2: char;
begin p:=Pred(Lo); chStart:=0;
      if Hi>Lo then
      repeat inc(p);
            if Body_^[p]=ch
            then
            begin if chStart=0 then chStart:=p;
            end
            else
            begin if chStart<>0
                  then
                  begin if p>chStart+1
                        then Move(Body_^[p],Body_^[chStart+1],Hi-p+1);
                        dec(Hi,p-chStart-1);
                        p:=chStart+1;
                        chStart:=0;

                  end
            end
      until p>=Hi;
      if (p=Hi)
      and (Body_^[p]=ch)
      and (chStart>0)
      and (p>chStart+1)
      then Hi:=chStart
end;

procedure chain.ExpandTo(newLength: word);
var tmpF: file of aCharVector;
var k,OldSize,i,p: word; tmpS_: aCharVector_;
    vector: aCharVector;
begin OldSize:=Size;
      if NewLength<OldSize then (* overflow *) ChainFail:=Overflow
      else
begin
      Size:=(newLength+Granularity-1) div Granularity * Granularity;
      (**
      while Size<newLength do Inc(Size,Granularity);
      ***)
      if  (MaxAvail>=Size) then
      begin GetMem(tmpS_,Size);
            Move(Body_^,tmpS_^,OldSize);
            FreeMem(Body_,OldSize);
            Body_:=tmpS_
      end
      else  (* try swapping to disk *)
      begin Assign(tmpF,'~~TMP~~.$$$');
            {$I-} Rewrite(tmpF); {$I+}
            if IOResult<>0 then ChainFail:=DiskSpace
            else
            begin
               k:=OldSize div Granularity; p:=1;
               for i:=1 to k do
               begin move(Body_^[p],vector,Granularity);
                  system.write(tmpF,vector);
                  inc(p,Granularity)
               end;
               FreeMem(Body_,OldSize);
               if (MaxAvail<Size)
               then ChainFail:=Memory
               else
               begin
                  GetMem(Body_,Size);
                  p:=1; Reset(tmpF);
                  for i:=1 to k do
                  begin system.read(tmpF,vector);
                        move(vector,Body_^[p],Granularity);
                        inc(p,Granularity)
                  end
               end
            end;
            {$I-} Close(tmpF); Erase(tmpF); {$I+}
      end;
      if ChainFail<>0 then Halt(1)
end
end;

procedure chain.read(terminator: char);
var c: char;
begin Clear;
      repeat system.read(c);
             if c<>terminator then AppendString(c);
      until c=terminator;
      if (terminator=#10) and (Body_^[Hi]=#13) then dec(Hi)
end;

procedure chain.AppendInput(terminator: char);
var c: char;
begin
      repeat system.read(c);
             if c<>terminator then AppendString(c);
      until c=terminator;
      if (terminator=#10) and (Body_^[Hi]=#13) then dec(Hi)
end;

procedure chain.write;
var i: word;
begin for i:=Lo to Hi do system.write(Body_^[i]);
end;

begin ChainFail:=0;
end.