Unit Long2;
Interface uses chain2;

type

   LongText= object
        f: file ;
        BufferSize: word;
        buffer,auxBuffer: chain;
        numRead,numWritten,minWidth,maxWidth: word;
        hyphen: char;
        lastEol,EolPos,PosInLine,lineLength,maxSegment,spacePos: word;
        ReadSofar,WrittenSoFar,FileSize: longint;
        EolFound,findSpace,formatted: boolean;
     end;
procedure LTAssign(var f: LongText; fn: string; bufSize: word);
procedure LTReset(var f: LongText);
procedure LTRewrite(var f: LongText);
procedure LTAppend(var f: LongText);

procedure LTClose(var f: LongText);
procedure LTReadChain(var f: LongText; var ch: chain);
function LTEof(var f: LongText): boolean;
procedure LTReadString(var f: LongText; var st:string);
procedure LTWriteChain(var f: LongText; var ch: chain);
procedure LTWriteString(var f: LongText; st: string);
procedure LTWriteChar(var f: LongText; ch: char);
procedure LTWriteLn(var f: LongText);
procedure LTSetFormat(var f: LongText;
                        minimumWidth,
                        maximumWidth: word);
procedure LTWriteSubchain(var f: LongText; var ch: chain; start,lgth: word);


Implementation
procedure LTAssign(var f: LongText; fn: string; bufSize: word);
begin FillChar(f,SizeOf(LongText),0);
      system.Assign(f.f,fn);
      f.buffer.Init(' ');
      f.buffer.SetGranularity(bufSize);
      f.auxBuffer.Init(' ');
      f.auxBuffer.SetGranularity(bufSize);
      f.BufferSize:=bufSize;
      f.PosInLine:=1;
      f.hyphen:='-';
      f.findSpace:=true
end;

procedure LTReset(var f: LongText);
begin
      system.Reset(f.f,1);
      f.FileSize:=system.FileSize(f.f);
end;

procedure LTAppend(var f: LongText);
begin LTReset(f);
      Seek(f.f,f.FileSize);
end;
procedure LTRewrite(var f: LongText);
begin
      system.rewrite(f.f,1);
end;

procedure LTClose(var f: LongText);
begin Close(f.f);
      f.Buffer.ReInit('');
      f.auxBuffer.ReInit('');
end;

function LTEof(var f: LongText):  boolean;
begin LTEof:=f.ReadSoFar>=f.FileSize;
end;

procedure LTReadChain(var f: LongText; var ch: chain);
var i: word; EolFound,CtrlZFound,EofFound: boolean;
begin
   ch.Clear; EolFound:=false; CtrlZFound:=false; EofFound:=false;
   repeat
      if f.numread=f.LastEol then
      begin
        f.lastEol:=0;
        f.buffer.Lo:=1;
        BlockRead(f.f,f.buffer.Body_^,f.BufferSize,f.numread);
      end;
      (* FindEol *);
      i:=f.buffer.Lo-1;
      repeat inc(i);
           case f.buffer.Body_^[i] of
              #10: begin EolFound:=true; EofFound:=true end;
              ^Z:  begin CtrlZFound:=true; EofFound:=true end;
           end
      until (EofFound) or (i=f.numread);
      f.buffer.Hi:=i;
      inc(f.ReadSoFar,f.buffer.length);
      ch.AppendChain(f.buffer);
      if EolFound then dec(ch.Hi,2) (* remove the end-of-line CRLF *)
      else if CtrlZFound then dec(ch.Hi,1); (* remove ^Z *)
      f.lastEol:=f.buffer.Hi;
   until EolFound or LTEof(f);
   f.buffer.Lo:=f.buffer.Hi+1;
end;

procedure LTReadString(var f: LongText; var st:string);
var size: word; i: byte;
begin
   LTReadChain(f,f.auxBuffer);
   size:=SizeOf(st)-1;
   if f.auxBuffer.Length<size then size:=f.auxBuffer.Length;
   Move(f.auxBuffer.Body_^,st[1],size);
   st[0]:=char(size);
end;

procedure LTWriteChainFormatted(var f: LongText; var ch: chain);
const HardHyphen='-';
var segment,leftInChain: word;

begin
  if ch.Length=0 then
  begin if f.lineLength>0 then LTWriteln(f);
        LTWriteln(f);
  end
  else
  begin if ch.Body_^[ch.Hi]<>' 'then ch.AppendString(' ');
      leftInChain:=ch.Length;
      repeat
        if f.findSpace then f.SpacePos:=ch.PosChar(' ');
        segment:=f.spacePos-ch.Lo+1;
        if segment<=f.maxSegment then
        begin BlockWrite(f.f,ch.Body_^[ch.Lo],segment);
              dec(f.maxSegment,segment);
              inc(f.lineLength,segment);
              inc(ch.Lo,segment);
              dec(leftInChain,segment);
              f.findSpace:=true
        end else
        if f.lineLength>=f.minWidth then
        begin LTWriteln(f);
              f.findSpace:=false;
        end
        else
        begin segment:=f.maxSegment-1;
              BlockWrite(f.f,ch.Body_^[ch.Lo],segment);
              LTWriteChar(f,f.hyphen);
              LTWriteLn(f);
              inc(ch.Lo,segment);
              dec(leftInChain,segment);
              f.findSpace:=false
        end
      until leftInChain=0
  end;
end;


procedure LTWriteChain(var f: LongText; var ch: chain);
begin if f.formatted then LTWriteChainFormatted(f,ch)
      else BlockWrite(f.f,ch.Body_^[ch.Lo],ch.Length,f.numWritten);
end;

procedure LTWriteString(var f: LongText; st: string);
begin if f.formatted then begin f.auxBuffer.CopyString(st);
                     LTWriteChainFormatted(f,f.auxBuffer);
             end
      else BlockWrite(f.f,st[1],length(st),f.numWritten);
end;

procedure LTWriteChar(var f: LongText; ch: char);
begin BlockWrite(f.f,ch,1,f.numWritten);
end;

procedure LTWriteLn(var f: LongText);
const st: string[2]=#13#10;
begin BlockWrite(f.f,st[1],length(st),f.numWritten);
      f.PosInLine:=1; f.lineLength:=0; f.maxSegment:=f.maxWidth
end;

procedure LTSetFormat(var f: LongText;
                        minimumWidth,
                        maximumWidth: word);
begin f.minWidth:=minimumWidth;
      f.maxWidth:=maximumWidth;
      if (f.minWidth=0) and (f.maxWidth=0) or (f.maxwidth<f.minwidth)
      then f.formatted:=false
      else f.formatted:=true;
      f.maxSegment:=f.maxWidth;
end;

procedure LTSetHyphen(var f: LongText; ch: char);
begin f.hyphen:=ch
end;

procedure LTWriteSubchain(var f: LongText; var ch: chain; start,lgth: word);
begin if start>ch.Hi then exit;
      if lgth>ch.Hi-start+1 then lgth:=ch.Hi-start+1;
      BlockWrite(f.f,ch.Body_^[start],lgth,f.numWritten);
end;


end.
